import math
from utils_navigation import *
# 

def directCPA(p0x, p0y, up, bp, q0x, q0y, vq, bq):
    # Computes the time and distance CPA
    # 

    # Input:
    # p0x, p0y - initial coordinates of the reference ship (yards)
    # up - speed of the reference ship (in knots)
    # bp - heading of the reference ship (in degrees)
    # q0x, q0y - initial coordinates of own ship (yards)
    # vq - speed of own ship (in knots)
    # bq - heading of own ship (in degrees)

    # Returns:
    # t - speed at which maneuver should be performed
    # distCPA - course at which maneuver should be performed

    # Convert angles to radians
    bp_rad = math.radians(bp)
    bq_rad = math.radians(bq)

    # Convert speeds to yards/min
    up = up*2000/60
    vq = vq*2000/60

    w0x = p0x - q0x
    w0y = p0y - q0y

    w0_squared = w0x**2 + w0y**2
    #print "w0x = ", w0x, "w0y = ", w0y, "w0_squared = ", w0_squared
    
    ux = up*math.cos(bp_rad)
    uy = up*math.sin(bp_rad)
    if ux < 1.0e-5:
        ux = 0
    if uy < 1.0e-5:
        uy = 0
    
    #print "ux = ", ux, "uy = ", uy
    

    vx = vq*math.cos(bq_rad)
    vy = vq*math.sin(bq_rad)
    if vx < 1.0e-5:
        vx = 0
    if vy < 1.0e-5:
        vy = 0
    #print "vx = ", vx, "vy = ", vy

    u_min_v_squared = (ux-vx)**2 + (uy-vy)**2
    if u_min_v_squared == 0:
        return 0, math.sqrt(w0_squared)

    w0x_uv = (w0x * (ux - vx) + w0y * (uy - vy))

    #print " u_min_v_squared = ", u_min_v_squared, " w0x_uv = ", w0x_uv, " w0_squared = ", w0_squared
    
    t = - w0x_uv/u_min_v_squared
    dist_CPA =  u_min_v_squared * t**2 + 2 * w0x_uv * t + w0_squared

    if  dist_CPA < 1.0e-5:
        dist_CPA = 0
        
    if dist_CPA < 0:
        return 0, t
    
    dist_CPA = math.sqrt(dist_CPA)
    dist_CPA = dist_CPA
    #print "t = ", t, "distCPA = ", dist_CPA
    return (t, dist_CPA)


def inverseCPA(p0x, p0y, up, bp, q0x, q0y, vq, bq, cpa_dist, s):
    # Computes the speed and angle of a boat that would increase the CPA
    # to cpa_dist

    # Input:
    # p0x, p0y - initial coordinates of the reference ship (yards)
    # up - speed of the reference ship (in knots)
    # bp - heading of the reference ship (in degrees)
    # q0x, q0y - initial coordinates of own ship (yards)
    # vq - speed of own ship (in knots)
    # bq - heading of own ship (in degrees)
    # s - min speed desired for the maneuver (knots)

    # Returns:
    # v - speed at which maneuver should be performed
    # b - course at which maneuver should be performed


    # Convert from trig. coordinates to this function's coordinate system
    bp = NormalizeHeading((360 - bp) + 90)
    bq = NormalizeHeading((360 - bq) + 90)

    # Convert angles to radians
    bp_rad = math.radians(bp)
    bq_rad = math.radians(bq)

    # Convert speeds to yards/min
    #up = up*2000/60
    #vq = vq*2000/60
    #s = s*2000/60


    w0x = p0x - q0x
    w0y = p0y - q0y
    w0_squared = w0x**2 + w0y**2

    a = cpa_dist**2 - w0_squared
    b = a + w0x**2
    c = a + w0y**2
    d = 2*w0x*w0y

    ux = up*math.cos(bp_rad)
    uy = up*math.sin(bp_rad)

    vx = math.cos(bq_rad)
    vy = math.sin(bq_rad)
    
    e = b * (ux**2) + c * (uy**2) + d * ux * uy
    m = b * vx**2 + c * vy**2 + d * vx * vy
    n = - (2*b*ux*vx + 2*c*uy*vy + d*ux*vy + vx*uy)

    delta = n**2 - 4*m*e

    if delta < 0:
        print "No solution to the problem exists: have to change heading"
        return (0, 0)

    v1 = (-n + math.sqrt(delta))/(2*m)
    v2 = (-n - math.sqrt(delta))/(2*m)

    if v1 <= 0 and v2 <= 0:
        # return a bearing for a speed of zero
        delta1 = d**2 - 4 *c*b
        if delta1 <=0:
            b = NormalizeHeading((360 - bq) + 90)
            v = 5
            #print "Inverse CPA: v = ", 0, "b = ", b
            return (v, b)
        
        #print "delta1 = ", delta1
        t1 = math.atan2 (-d+math.sqrt(delta1), 2*c)
        t2 = math.atan2 (-d-math.sqrt(delta1), 2*c)

        #print t1*180/math.pi, t2*180/math.pi
        b = t1*180/math.pi
        b = NormalizeHeading((360 - b) + 90)
        #print "Inverse CPA: v = ", 0, "b = ", b
        return (0, b)
    else:
        if v1 > 0:
            #v = (v1 * 60)/2000
            v = v1
            b =  NormalizeHeading((360 - bq) + 90)
        elif v2 > 0:
            #v = (v1 * 60)/2000
            v = v2
            b =  NormalizeHeading((360 - bq) + 90)
        else:
            print "Error: should not get here"
            v = 0
            b = 0
    #print "Inverse CPA: v = ", v, "b = ", b
    return (v, b)
