#
# @(#) executive.py 2005/07/06
# @author Copyright (c) 2005 The University of Nevada, Reno. All
# rights reserved.
#
# This software is provided "AS IS," without a warranty of any
# kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
# WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
# EXCLUDED. THE UNIVERSITY OF NEVADA, RENO AND ITS LICENSORS SHALL NOT BE
# LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING,
# MODIFYING OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO
# EVENT WILL THE UNIVERSITY OF NEVADA, RENO OR ITS LICENSORS BE LIABLE FOR
# ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
# CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND
# REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR
# INABILITY TO USE SOFTWARE, EVEN IF THE UNIVERSITY OF NEVADA, RENO HAS BEEN
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
#
#TODO: Fix Header
"""
Parent class of all behaviors.

Behavior has no functionality of its own. It defines methods and data common to
all behaviors. It should be considered abstract, but python doesn't do that.
"""
import wx
from cat.Interface import sensors

class Behavior:
    """
    The parent class of all behaviors
    
    Because of their use in networks, variables that need to instantiated later
    need to be registered through special methods. Variables that are just used
    internally can be declared normally.
    
    If you want to create your own primitive behavior, then you should inherit
    from primitive behavior and not this.
    """
    
    def __init__(self, sens = sensors.guiSensors):
        """
        Creates an empty behavior.
        """
        self.weight = 0
        self.heading = 0
        self.speed = 0
        self.sens = sens
        self.vars = {}
        self.preTriggers = []
        self.postTriggers = []
        self.active = False

    def activate(self):
        from cat.entity import EntityManager
        if EntityManager.currentState:
            EntityManager.currentState.lastviewed.append(self.type)
        self.active = True

    def deActivate(self):
        from cat.entity import EntityManager
        if EntityManager.currentState and EntityManager.currentState.lastviewed.count(self.type) > 0:
            EntityManager.currentState.lastviewed.remove(self.type)
        self.active = False
        
    def GetAction(self):
        """
        Returns the current actuator values.
        """
        return (self.weight, self.heading, self.speed)
        
    def Tick(self, time, dtime):
        """
        Returns the current actuator values.
        
        Does no computation on its own. This function should be overriden.
        """
        return self.GetAction()

    def AddBool(self, name, val):
        """
        Registers a boolean.
        
        val (bool): default value
        """
        self.vars[name] = (val, "bool", False, True)

    def AddInt(self, name, val, min, max):
        """
        Registers an int
        
        val (int): default value
        min (int): minimum value
        max (int): maximum value
        """
        self.vars[name] = (val, "int", min, max)

    def AddFloat(self, name, val, min, max):
        """
        Registers a float
        
        val (float): default value
        min (float): minimum value
        max (float): maximum value
        """    
        self.vars[name] = (val, "float", min, max)

    def AddVector(self, name, val, min, max):
        self.vars[name] = (val, "vector", min, max)

    def GetVar(self, name):
        """
        Gets the value of a registered variable
        
        The vars are all in a dictionary but are tuples. This unwarps the tuple
        to return the right value.
        """
        return self.vars[name][0]

    def SetVar(self, name, val):
        """
        Sets the values of a Var
        
        A convinience wrapper for setting the value of a var inside the tuple
        dictionary.
        """
        old, type, min, max = self.vars[str(name)]
        self.vars[str(name)] = (val, type, min, max)

    def HasVar(self, name):
        """
        Sets the values of a Var
        
        A convinience wrapper for setting the value of a var inside the tuple
        dictionary.
        """
        return self.vars.has_key(str(name))

    def IsActive(self):
        return self.active

    def isActive(self):
        return self.active
    
    def checkPostconditions(self):
        return False

    def checkPreconditions(self):
        return False
