#
# @(#) executive.py 2005/07/06
# @author Copyright (c) 2005 The University of Nevada, Reno. All
# rights reserved.
#
# This software is provided "AS IS," without a warranty of any
# kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
# WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
# EXCLUDED. THE UNIVERSITY OF NEVADA, RENO AND ITS LICENSORS SHALL NOT BE
# LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING,
# MODIFYING OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO
# EVENT WILL THE UNIVERSITY OF NEVADA, RENO OR ITS LICENSORS BE LIABLE FOR
# ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
# CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND
# REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR
# INABILITY TO USE SOFTWARE, EVEN IF THE UNIVERSITY OF NEVADA, RENO HAS BEEN
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
#
#TODO: Fix Header
"""
Further defines methods and data used by primitive behavior GUIs. 
Contains the PanelGroup for mutual exclusive behaviors.
"""
import wx
from behavior import *
from cat.Interface import sensors, important
from common.defs import *
from cat.entity import EntityManager

class PanelGroup:
    """
    All Panels added to this group are mutually exclusive and only one can be
    active at a time.
    """
    panels = []
    def AddPanel(self, panel):
        """
        Adds a panel to the group.
        """
        self.panels.append(panel)
        panel.panelGroup = self
        
    def SwapPanel(self, activePanel):
        """
        Every behavior that is the active behavior is deactivated.
        """
        for panel in self.panels:
            if panel != activePanel:
                panel._OnDeactivate(None)
                
class PrimitiveBehavior(Behavior):
    """
    Parent class for primitive behaviors.
    
    It's empty right now. It's mostly for clarity in the inheritance chain.
    """
    def __init__(self, sens=sensors.guiSensors, type='none'):
        Behavior.__init__(self, sens)
        self.AddBool('avoidentity', False)
        self.AddBool('avoidland', False)
        self.AddFloat('epPanicRange', 100.0, 0.0, 1000.0)
        self.AddFloat('epEffectiveRange', 150.0, 0.0, 1000.0)
        self.AddFloat('epDesiredSpeed', 10.0, 0.0, 30.0)
        self.AddFloat("lpRange", 1000.0,0.0,10000.0)
        self.AddFloat("lpSpeed", 0.0,0.0,28.0)
        self.type = type

    def getBehaviorType(self):
        return self.type

class PrimitiveBehaviorPanel(wx.Panel):
    """
    A wxPanel for setting up a PrimitiveBehavior GUI.
    
    There is a lot of inner plumbing to set up a GUI and turn it on and off.
    This class tries to handle as much of that as possible.
    """
    
    def __init__(self, parent, panelID, buttonID, name, stop=None, type='none'):
        """
        Sets up a simple GUI.
        
        Really simple. Like one button that turns the panel on and off. When you
        create your own GUI, make your own sizer and add it to self.sizer. Also,
        you'll want to set self.beh to the behavior your writing the GUI for.
        """
        wx.Panel.__init__(self, parent, panelID)
        self.ID = buttonID
        self.name = name
        self.SetBackgroundColour(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DLIGHT))
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)
        self.activate = wx.Button(self, self.ID, "Activate")
        self.sizer.Add(self.activate, border=5, flag=wx.TOP|wx.LEFT)
        wx.EVT_BUTTON(self.activate, self.ID, self._OnActivate)
        self.bset = None
        self.active = False
        self.stopPanel = stop
        self.type = type
    
    def _OnActivate(self,event):
        """
        Private function for handling activation background work.
        """
        if EntityManager.currentState:
            p = self.GetParent()
            p.SetPageText(self.pageNum, self.name+" [ON]")
            self.activate.SetLabel("Deactivate")
            wx.EVT_BUTTON(self.activate, self.ID, self._OnDeactivate)
            self.active = True
            beh = EntityManager.currentState.bset.getList(self.type)[0]
            beh.active = self.active
            self.OnActivate(event)
            if self.stopPanel:
                self.stopPanel._OnDeactivate(None)
            important.ImportantEvent()

    def _OnDeactivate(self,event):
        """
        Private function for handing deactivation background work.
        """
        if EntityManager.currentState:
            p = self.GetParent()
            p.SetPageText(self.pageNum, self.name+" [off]")
            self.activate.SetLabel("Activate")
            wx.EVT_BUTTON(self.activate, self.ID, self._OnActivate)
            self.active = False
            beh = EntityManager.currentState.bset.getList(self.type)[0]
            beh.active = self.active
            self.OnDeactivate(event)
            if exists(beh.sens.GetOwnEntity()):
                sel(beh.sens.GetOwnEntity())
                sf('desiredSpeed', 0.0)
            important.ImportantEvent()
    
    def OnActivate(self, event):
        """
        Empty method called by _OnActivate()
        
        If you require special routines on activation, override this function to
        do that.
        """
        pass
        
    def OnDeactivate(self, event):
        """
        Empty method called by _OnDeactivate()
        
        If you require special routines on deactivation, override this function to
        do that.
        """    
        pass
        
    def Tick(self, time, dtime):
        """
        Returns the actuator values from the attached behavior.
        """
        return (0.0, 0.0)
            
    def IsActive(self):
        return self.active

    def _refresh(self):
        if EntityManager.currentState:
            beh = EntityManager.currentState.bset.getList(self.type)[0]
            if beh.active:
                self._OnActivate(None)
            else:
                self._OnDeactivate(None)
            self.refresh()

    def refresh(self):
        pass

