import math
import time
import behaviorLink

#uses A* to sequentially lay out lines
def makeLineLayout(behaviorShapes, shapePairList, canvasSize):
    #make the shape list to pass to the expand function
    shapeList = []
    for behavior, shape in behaviorShapes.items():
        shapeList.append(shape)

    #get the dimensions of the canvas
    canvasw, canvash = canvasSize

    #control variables:
    lineStep = 15               #the distance between vertices in a line
    thStep = math.pi/5          #the step in radians taken between attach points
    numDirs = 12                #the number of directions expanded into
    lineCost = lineStep * 2     #the cost to cross another line
    shapeCost = lineStep * 3    #the cost to cross a behaviorShape
    turnCost = lineStep / 2     #the cost to turn all the way around

    #traverse shapePairs to make lines
    lineList = []
    lastLink = None
    for shapePair in shapePairList:

        #initialize the stack with a 0 cost line
        lineStack = [Line(shapePair,lineStep,thStep)]
        finishedLine = None

        while len(lineStack) > 0 and not finishedLine:
            #get the line with lowest cost + estimate
            line = lineStack[0]
            minCost = line.cost + line.getEstimate()
            for stackLine in lineStack:
                cost = stackLine.cost + stackLine.getEstimate()
                if cost < minCost:
                    minCost = cost
                    line = stackLine

            #expand that line and remove it from the list
            newLines = line.expand(lineList,shapeList,numDirs,lineCost,shapeCost,turnCost)
            lineStack.remove(line)

            #add the new lines if they aren't duplicates of lines 
            #in the stack and they're in the canvas
            for newLine in newLines:
                if newLine.finished:
                    finishedLine = newLine
                duplicate = False
                x,y = newLine.vertices[-1]
                for line in lineStack:
                    if line.collides(x,y):
                        duplicate = True
                if not duplicate and x > 0 and x < canvasw and y > 0 and y < canvash:
                    lineStack.append(newLine)

        #if we found one append it to the overall list
        if finishedLine:
            lineList.append(finishedLine)
        else:
        #else error, this should really never happen and would take really long if it did
            fromShape, toShape = shapePair
            print "ERROR: unable to find a line to connect %s and %s" % (fromShape.behavior.bname,toShape.behavior.bname)

    return lineList

#holds info about one line
class Line:
    #make the basic line information
    def __init__(self, shapePair, lineStep, thStep, fromTh = None, toTh = None):
        #copy vars
        self.fromShape, self.toShape = shapePair
        self.lineStep = lineStep
        self.thStep = thStep
        self.fromTh = fromTh
        self.toTh = toTh

        #set local vars
        self.lastTh = 0
        self.finished = False
        self.cost = 0

        #get info about location and size of shapes
        tox = self.toShape.GetX()
        toy = self.toShape.GetY()
        tow = self.toShape.GetWidth()
        toh = self.toShape.GetHeight()

        fromx = self.fromShape.GetX()
        fromy = self.fromShape.GetY()
        fromw = self.fromShape.GetWidth()
        fromh = self.fromShape.GetHeight()

        #init vertices
        self.vertices = []

        #figure out angle between shapes
        dx = float(tox-fromx)
        dy = float(toy-fromy)
        th = math.acos(dx/math.hypot(dx,dy))
        if toy < fromy:
            th = th * -1

        #make the start vertex if it isn't provided and add it to vertices
        if not self.fromTh:
            self.fromTh = self.makeTh(th,self.fromShape.thList)
        startx = fromx + fromw/2 * math.cos(self.fromTh)
        starty = fromy + fromh/2 * math.sin(self.fromTh)
        self.startVertex = startx,starty
        self.vertices.append(self.startVertex)

        #make the stop vertex if it isn't provided and save it for later
        if not self.toTh:
            self.toTh = self.makeTh(th+math.pi,self.toShape.thList)
        stopx = tox + tow/2 * math.cos(self.toTh)
        stopy = toy + toh/2 * math.sin(self.toTh)
        self.stopVertex = stopx,stopy

    #deep copy line into self
    def deepCopy(self, line):
        self.fromShape = line.fromShape
        self.toShape = line.toShape
        self.lineStep = line.lineStep
        self.thStep = line.thStep
        self.fromTh = line.fromTh
        self.toTh = line.toTh
        self.finished = line.finished
        self.cost = line.cost
        self.stopVertex = line.stopVertex
        self.lastTh = line.lastTh

        self.vertices = []
        for vertex in line.vertices:
            self.vertices.append(vertex)

    #underestimating heuristic for A*
    def getEstimate(self):
        stopx, stopy = self.stopVertex
        x, y = self.vertices[-1]
        return math.hypot(stopx-x,stopy-y) * .99

    #expands the end vertex of this line using parameters from makeLineLayout
    def expand(self, lineList, shapeList, numDirs, lineCost, shapeCost, turnCost):
        newLines = []
        oldx,oldy = self.vertices[-1]
        for i in range(numDirs):
            #make new vertex
            th = float(i) / float(numDirs) * math.pi * 2
            newx = oldx + self.lineStep * math.cos(th)
            newy = oldy + self.lineStep * math.sin(th)
            #allocate line and append new vertex
            newLine = Line((self.fromShape,self.toShape),self.lineStep,self.thStep,self.fromTh,self.toTh)
            newLine.deepCopy(self)
            newLine.vertices.append((newx,newy))
            #calculate cost
            newLine.cost = newLine.cost + newLine.lineStep
            newLine.cost = newLine.cost + math.fabs(th-newLine.lastTh) / math.pi * turnCost
            newLine.lastTh = th
            for line in lineList:
                if line.collides(newx,newy):
                    newLine.cost = newLine.cost + lineCost
            for shape in shapeList:
                if shape.collides(newx,newy):
                    newLine.cost = newLine.cost + shapeCost
            #check if this line is finished
            newLine.checkFinished()
            #append
            newLines.append(newLine)
        return newLines

    #checks if this line has gotten to its goal and fixes the last vertex
    def checkFinished(self):
        stopx, stopy = self.stopVertex
        x,y = self.vertices[-1]
        if math.hypot(stopx-x,stopy-y) <= self.lineStep:
            self.vertices.append(self.stopVertex)
            self.finished = True

    #checks if the test point is within lineStep of any vertices in this line
    def collides(self, testx, testy):
        for x,y in self.vertices:
            if math.hypot(testx-x,testy-y) < self.lineStep:
                return True
        return False

    #simple list compare with tolerance
    #called by makeTh
    def listContains(self, n, list, tol):
        for num in list:
            if n >= num - tol and n <= num + tol:
                return True
        return False

    #finds a point to attach a line to a shape at by bouncing back and forth on the target
    def makeTh(self, targetTh, thList):
        th = targetTh
        sign = 1
        diff = 0

        while self.listContains(th,thList,self.thStep) and diff <= math.pi:
            if sign == -1:
                diff = diff + self.thStep
            sign = sign * -1
            th = targetTh + diff * sign

        if diff > math.pi:
            th = targetTh

        thList.append(th)
        return th



